@@ -21,6 +21,7 @@
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */
 
 #include "priv.h"
+#include <version.h>
 
 static void
 _print_version (FILE *stream, struct argp_state *state)
@@ -28,12 +29,17 @@ _print_version (FILE *stream, struct argp_state *state)
   if (argp_program_version)
     /* If this is non-zero, then the program's probably defined it, so let
        that take precedence over the default.  */
-    fprintf (stream, "%s\n", argp_program_version);
+    fputs (argp_program_version, stream);
   else
-    /* Construct a version using the standard diskfs variables.  */
-    fprintf (stream, "%s %d.%d (GNU %s)\n",
-	     diskfs_server_name, diskfs_major_version, diskfs_minor_version,
-	     HURD_RELEASE);
+    {
+      fprintf (stream, "%s %s ", diskfs_server_name, diskfs_server_version);
+      if (diskfs_extra_version[0])
+	fprintf (stream, "(%s) ", diskfs_extra_version);
+    }
+  /* And because diskfs is big and huge, put our information out too. */
+  fputs (STANDARD_HURD_VERSION (libdiskfs), stream);
+  
+  putc ('\n', stream);
 }
 
 void (*argp_program_version_hook) (FILE *stream, struct argp_state *state)
