@@ -709,23 +709,48 @@ argp_args_usage (const struct argp *argp, FILE *stream)
       argp_args_usage (*children++, stream);
 }
 
-/* Print the documentation for ARGP to STREAM.  Each separate bit of
-   documentation is preceded by a blank line.  */
-static void
-argp_doc (const struct argp *argp, FILE *stream)
+/* Print the documentation for ARGP to STREAM; if POST is false, then
+   everything preceeding a `\v' character in the documentation strings (or
+   the whole string, for those with none) is printed, otherwise, everything
+   following the `\v' character (nothing for strings without).  Each separate
+   bit of documentation is separated a blank line, and if PRE_BLANK is true,
+   then the first is as well.  If FIRST_ONLY is true, only the first
+   occurance is output.  Returns true if anything was output.  */
+static int
+argp_doc (const struct argp *argp, int post, int pre_blank, int first_only,
+	  FILE *stream)
 {
   const struct argp **children = argp->children;
   const char *doc = argp->doc;
+  int anything = 0;
+
   if (doc)
     {
-      putc ('\n', stream);
-      fputs (doc, stream);
+      char *vt = index (doc, '\v');
+
+      if (pre_blank && (vt || !post))
+	putc ('\n', stream);
+
+      if (vt)
+	if (post)
+	  fputs (vt + 1, stream);
+	else
+	  fwrite (doc, 1, vt - doc, stream);
+      else
+	if (! post)
+	  fputs (doc, stream);
       if (line_wrap_point (stream) > line_wrap_lmargin (stream))
 	putc ('\n', stream);
+
+      anything = 1;
     }
   if (children)
-    while (*children)
-      argp_doc (*children++, stream);
+    while (*children && !(first_only && anything))
+      anything |=
+	argp_doc (*children++, post, anything || pre_blank, first_only,
+		  stream);
+
+  return anything;
 }
 
 /* Output a usage message for ARGP to STREAM.  FLAGS are from the set
@@ -733,7 +758,7 @@ argp_doc (const struct argp *argp, FILE *stream)
 void argp_help (const struct argp *argp, FILE *stream,
 		unsigned flags, char *name)
 {
-  int first = 1;
+  int anything = 0;		/* Whether we've output anything.  */
   struct hol *hol = 0;
 
   if (! stream)
@@ -780,13 +805,18 @@ void argp_help (const struct argp *argp, FILE *stream,
       line_wrap_set_lmargin (stream, old_lm);
 
       putc ('\n', stream);
-      first = 0;
+      anything = 1;
     }
 
+  if (flags & ARGP_HELP_PRE_DOC)
+    anything |= argp_doc (argp, 0, 0, 1, stream);
+
   if (flags & ARGP_HELP_SEE)
     {
-      fprintf (stream, "Try `%s --help' for more information.\n", name);
-      first = 0;
+      fprintf (stream,
+	       "Try `%s --help' or `%s --usage' for more information.\n",
+	       name, name);
+      anything = 1;
     }
 
   if (flags & ARGP_HELP_LONG)
@@ -795,14 +825,23 @@ void argp_help (const struct argp *argp, FILE *stream,
       /* Print info about all the options.  */
       if (hol->num_entries > 0)
 	{
-	  if (! first)
+	  if (anything)
 	    putc ('\n', stream);
 	  hol_help (hol, stream);
-	  first = 0;
+	  anything = 1;
 	}
+    }
 
-      /* Finally, print any documentation strings at the end.  */
-      argp_doc (argp, stream);
+  if (flags & ARGP_HELP_POST_DOC)
+    /* Print any documentation strings at the end.  */
+    anything |= argp_doc (argp, 1, anything, 0, stream);
+
+  if ((flags & ARGP_HELP_BUG_ADDR) && argp_program_bug_address)
+    {
+      if (anything)
+	putc ('\n', stream);
+      fprintf (stream, "Report bugs to %s.\n", argp_program_bug_address);
+      anything = 1;
     }
 
   if (hol)
