@@ -74,6 +74,61 @@ register_fresh_stat (struct node *np, int *p)
   return ret;
 }
 
+/* Handle returned wcc information for various calls.  In protocol
+   version 2, this is just register_fresh_stat.  In version 3, it
+   checks to see if stat information is present too.  If this follows
+   an operation that we expect has modified the attributes, MOD should
+   be set.  (This unpacks the post_op_attr XDR type.)  */
+int *
+process_returned_stat (struct node *np, int *p, int mod)
+{
+  int attrs_exist;
+
+  if (protocol_version == 2)
+    return register_fresh_stat (np, p);
+  else
+    {
+      attrs_exist = ntohl (*p++);
+      if (attrs_exist)
+	p = register_fresh_stat (np, p);
+      else if (mod)
+	/* We know that our values are now wrong */
+	np->nn->stat_updated = 0;
+      return p;
+    }
+}
+
+
+/* Handle returned wcc information for various calls.  In protocol
+   version 2, this is just register_fresh_stat.  In version 3, it does
+   the wcc_data interpretation too.  If this follows an operation that
+   we expect has modified the attributes, MOD should be set. 
+   (This unpacks the wcc_data XDR type.) */
+int *
+process_wcc_stat (struct node *np, int *p, int mod)
+{
+  if (protocol_version == 2)
+    return register_fresh_stat (np, p);
+  else
+    {
+      int attrs_exist;
+
+      /* First the pre_op_attr */
+      attrs_exist = ntohl (*p++);
+      if (attrs_exist)
+	{
+	  /* Just skip them for now */
+	  p += 2 * sizeof (int); /* size */
+	  p += 2 * sizeof (int); /* mtime */
+	  p += 2 * sizeof (int); /* atime */
+	}
+      
+      /* Now the post_op_attr */
+      return process_returned_stat (np, p, mod);
+    }
+}
+
+
 /* Implement the netfs_validate_stat callback as described in
    <hurd/netfs.h>. */
 error_t
@@ -86,8 +141,8 @@ netfs_validate_stat (struct node *np, struct netcred *cred)
   if (mapped_time->seconds - np->nn->stat_updated < stat_timeout)
     return 0;
 
-  p = nfs_initialize_rpc (NFSPROC_GETATTR, (struct netcred *) -1,
-			  0, &rpcbuf, np, -1);
+  p = nfs_initialize_rpc (NFSPROC_GETATTR (protocol_version),
+			  (struct netcred *) -1, 0, &rpcbuf, np, -1);
   p = xdr_encode_fhandle (p, &np->nn->handle);
   
   err = conduct_rpc (&rpcbuf, &p);
@@ -112,14 +167,20 @@ netfs_attempt_chown (struct netcred *cred, struct node *np,
   void *rpcbuf;
   error_t err;
     
-  p = nfs_initialize_rpc (NFSPROC_SETATTR, cred, 0, &rpcbuf, np, gid);
+  p = nfs_initialize_rpc (NFSPROC_SETATTR (protocol_version),
+			  cred, 0, &rpcbuf, np, gid);
   p = xdr_encode_fhandle (p, &np->nn->handle);
   p = xdr_encode_sattr_ids (p, uid, gid);
+  if (protocol_version == 3)
+    *p++ = 0;			/* guard_check == 0 */
   
   err = conduct_rpc (&rpcbuf, &p);
-
   if (!err)
-    register_fresh_stat (np, p);
+    {
+      err = nfs_error_trans (ntohl (*p++));
+      if (!err || protocol_version == 3)
+	p = process_wcc_stat (np, p, !err);
+    }
 
   free (rpcbuf);
   
@@ -180,17 +241,21 @@ netfs_attempt_chmod (struct netcred *cred, struct node *np,
 	}
     }
 
-  p = nfs_initialize_rpc (NFSPROC_SETATTR, cred, 0, &rpcbuf, np, -1);
+  p = nfs_initialize_rpc (NFSPROC_SETATTR (protocol_version),
+			  cred, 0, &rpcbuf, np, -1);
   p = xdr_encode_fhandle (p, &np->nn->handle);
   p = xdr_encode_sattr_mode (p, mode);
+  if (protocol_version == 3)
+    *p++ = 0;			/* guard check == 0 */
   
   err = conduct_rpc (&rpcbuf, &p);
   if (!err)
-    err = nfs_error_trans (ntohl (*p++));
+    {
+      err = nfs_error_trans (ntohl (*p++));
+      if (!err || protocol_version == 3)
+	p = process_wcc_stat (np, p, !err);
+    }
   
-  if (!err)
-    register_fresh_stat (np, p);
-
   free (rpcbuf);
   return err;
 }
@@ -214,16 +279,20 @@ netfs_attempt_utimes (struct netcred *cred, struct node *np,
   void *rpcbuf;
   error_t err;
     
-  p = nfs_initialize_rpc (NFSPROC_SETATTR, cred, 0, &rpcbuf, np, -1);
+  p = nfs_initialize_rpc (NFSPROC_SETATTR (protocol_version),
+			  cred, 0, &rpcbuf, np, -1);
   p = xdr_encode_fhandle (p, &np->nn->handle);
   p = xdr_encode_sattr_times (p, atime, mtime);
+  if (protocol_version == 3)
+    *p++ = 0;			/* guard check == 0 */
   
   err = conduct_rpc (&rpcbuf, &p);
   if (!err)
-    err = nfs_error_trans (ntohl (*p++));
-  
-  if (!err)
-    register_fresh_stat (np, p);
+    {
+      err = nfs_error_trans (ntohl (*p++));
+      if (!err || protocol_version == 3)
+	p = process_wcc_stat (np, p, !err);
+    }
 
   free (rpcbuf);
   return err;
@@ -239,17 +308,21 @@ netfs_attempt_set_size (struct netcred *cred, struct node *np,
   void *rpcbuf;
   error_t err;
   
-  p = nfs_initialize_rpc (NFSPROC_SETATTR, cred, 0, &rpcbuf, np, -1);
+  p = nfs_initialize_rpc (NFSPROC_SETATTR (protocol_version),
+			  cred, 0, &rpcbuf, np, -1);
   p = xdr_encode_fhandle (p, &np->nn->handle);
   p = xdr_encode_sattr_size (p, size);
+  if (protocol_version == 3)
+    *p++ = 0;			/* guard_check == 0 */
   
   err = conduct_rpc (&rpcbuf, &p);
   if (!err)
-    err = nfs_error_trans (ntohl (*p++));
+    {
+      err = nfs_error_trans (ntohl (*p++));
+      if (!err || protocol_version == 3)
+	p = process_wcc_stat (np, p, !err);
+    }
   
-  if (!err)
-    register_fresh_stat (np, p);
-
   free (rpcbuf);
   return err;
 }
@@ -264,7 +337,7 @@ netfs_attempt_statfs (struct netcred *cred, struct node *np,
   void *rpcbuf;
   error_t err;
     
-  p = nfs_initialize_rpc (NFSPROC_STATFS, cred, 0, &rpcbuf, np, -1);
+  p = nfs_initialize_rpc (NFS2PROC_STATFS, cred, 0, &rpcbuf, np, -1);
   p = xdr_encode_fhandle (p, &np->nn->handle);
   
   err = conduct_rpc (&rpcbuf, &p);
@@ -320,6 +393,7 @@ netfs_attempt_read (struct netcred *cred, struct node *np,
   size_t trans_len;
   error_t err;
   size_t amt, thisamt;
+  int eof;
   
   for (amt = *len; amt;)
     {
@@ -327,40 +401,49 @@ netfs_attempt_read (struct netcred *cred, struct node *np,
       if (thisamt > read_size)
 	thisamt = read_size;
 
-      p = nfs_initialize_rpc (NFSPROC_READ, cred, 0, &rpcbuf, np, -1);
+      p = nfs_initialize_rpc (NFSPROC_READ (protocol_version),
+			      cred, 0, &rpcbuf, np, -1);
       p = xdr_encode_fhandle (p, &np->nn->handle);
       *p++ = htonl (offset);
       *p++ = htonl (thisamt);
-      *p++ = 0;
+      if (protocol_version == 2)
+	*p++ = 0;
   
       err = conduct_rpc (&rpcbuf, &p);
       if (!err)
-	err = nfs_error_trans (ntohl (*p++));
-  
-      if (err)
 	{
-	  free (rpcbuf);
-	  return err;
-	}
-      
-      p = register_fresh_stat (np, p);
-      
-      trans_len = ntohl (*p++);
-      if (trans_len > thisamt)
-	trans_len = thisamt;	/* ??? */
-      
-      bcopy (p, data, trans_len);
-      free (rpcbuf);
+	  err = nfs_error_trans (ntohl (*p++));
 
-      data += trans_len;
-      offset += trans_len;
-      amt -= trans_len;
+	  if (!err || protocol_version == 3)
+	    p = process_returned_stat (np, p, !err);
+
+	  if (err)
+	    {
+	      free (rpcbuf);
+	      return err;
+	    }
       
-      /* If we got a short count, that means we're all done */
-      if (trans_len < thisamt)
-	{
-	  *len -= amt;
-	  return 0;
+	  trans_len = ntohl (*p++);
+	  if (trans_len > thisamt)
+	    trans_len = thisamt;	/* ??? */
+
+	  if (protocol_version == 3)
+	    eof = ntohl (*p++);
+	  else
+	    eof = (trans_len < thisamt);
+	  
+	  bcopy (p, data, trans_len);
+	  free (rpcbuf);
+
+	  data += trans_len;
+	  offset += trans_len;
+	  amt -= trans_len;
+      
+	  if (eof)
+	    {
+	      *len -= amt;
+	      return 0;
+	    }
 	}
     }
   return 0;
@@ -376,6 +459,7 @@ netfs_attempt_write (struct netcred *cred, struct node *np,
   void *rpcbuf;
   error_t err;
   size_t amt, thisamt;
+  size_t count;
   
   for (amt = *len; amt;)
     {
@@ -383,32 +467,80 @@ netfs_attempt_write (struct netcred *cred, struct node *np,
       if (thisamt > write_size)
 	thisamt = write_size;
       
-      p = nfs_initialize_rpc (NFSPROC_WRITE, cred, thisamt, &rpcbuf, np, -1);
+      p = nfs_initialize_rpc (NFSPROC_WRITE (protocol_version),
+			      cred, thisamt, &rpcbuf, np, -1);
       p = xdr_encode_fhandle (p, &np->nn->handle);
-      *p++ = 0;
+      if (protocol_version == 2)
+	*p++ = 0;
       *p++ = htonl (offset);
-      *p++ = 0;
+      if (protocol_version == 2)
+	*p++ = 0;
+      if (protocol_version == 3)
+	*p++ = htonl (FILE_SYNC);
       p = xdr_encode_data (p, data, thisamt);
   
       err = conduct_rpc (&rpcbuf, &p);
       if (!err)
-	err = nfs_error_trans (ntohl (*p++));
-  
+	{
+	  err = nfs_error_trans (ntohl (*p++));
+	  if (!err || protocol_version == 3)
+	    p = process_wcc_stat (np, p, !err);
+	  if (!err)
+	    {
+	      if (protocol_version == 3)
+		{
+		  count = ntohl (*p++);
+		  p++;		/* ignore COMMITTED */
+		  /* ignore verf for now */
+		  p += NFS3_WRITEVERFSIZE / sizeof (int);		  
+		}
+	      else
+		/* assume it wrote the whole thing */
+		count = thisamt;
+	      
+	      free (rpcbuf);
+	      amt -= count;
+	      data += count;
+	      offset += count;
+	    }
+	}
+      
       if (err)
 	{
 	  *len = 0;
 	  free (rpcbuf);
 	  return err;
 	}
-      register_fresh_stat (np, p);
-      free (rpcbuf);
-      amt -= thisamt;
-      data += thisamt;
-      offset += thisamt;
     }
   return 0;
 }
 
+/* See if NAME exists in DIR for CRED.  If so, return EEXIST.  */
+error_t
+verify_nonexistent (struct netcred *cred, struct node *dir,
+		    char *name)
+{
+  int *p;
+  void *rpcbuf;
+  error_t err;
+  
+  assert (protocol_version == 2);
+
+  p = nfs_initialize_rpc (NFSPROC_LOOKUP (protocol_version),
+			  cred, 0, &rpcbuf, dir, -1);
+  p = xdr_encode_fhandle (p, &dir->nn->handle);
+  p = xdr_encode_string (p, name);
+
+  err = conduct_rpc (&rpcbuf, &p);
+  if (!err)
+    err = nfs_error_trans (ntohl (*p++));
+  
+  if (!err)
+    return EEXIST;
+  else
+    return 0;
+}
+
 /* Implement the netfs_attempt_lookup callback as described in
    <hurd/netfs.h>. */
 error_t
@@ -419,21 +551,34 @@ netfs_attempt_lookup (struct netcred *cred, struct node *np,
   void *rpcbuf;
   error_t err;
   
-  p = nfs_initialize_rpc (NFSPROC_LOOKUP, cred, 0, &rpcbuf, np, -1);
+  p = nfs_initialize_rpc (NFSPROC_LOOKUP (protocol_version),
+			  cred, 0, &rpcbuf, np, -1);
   p = xdr_encode_fhandle (p, &np->nn->handle);
   p = xdr_encode_string (p, name);
   
   mutex_unlock (&np->lock);
 
   err = conduct_rpc (&rpcbuf, &p);
-  if (!err)
-    err = nfs_error_trans (ntohl (*p++));
-  
   if (!err)
     {
-      *newnp = lookup_fhandle (p);
-      p += NFS_FHSIZE / sizeof (int);
-      register_fresh_stat (*newnp, p);
+      err = nfs_error_trans (ntohl (*p++));
+      if (!err)
+	{
+	  p = lookup_fhandle (p, newnp);
+	  p = process_returned_stat (*newnp, p, 1);
+	}
+      if (err)
+	*newnp = 0;
+      if (protocol_version == 3)
+	{
+	  if (*newnp)
+	    mutex_unlock (&(*newnp)->lock);
+	  mutex_lock (&np->lock);
+	  p = process_returned_stat (np, p, 0);
+	  mutex_unlock (&np->lock);
+	  if (*newnp)
+	    mutex_lock (&(*newnp)->lock);
+	}
     }
   else
     *newnp = 0;
@@ -453,7 +598,8 @@ netfs_attempt_mkdir (struct netcred *cred, struct node *np,
   void *rpcbuf;
   error_t err;
   
-  p = nfs_initialize_rpc (NFSPROC_MKDIR, cred, 0, &rpcbuf, np, -1);
+  p = nfs_initialize_rpc (NFSPROC_MKDIR (protocol_version),
+			  cred, 0, &rpcbuf, np, -1);
   p = xdr_encode_fhandle (p, &np->nn->handle);
   p = xdr_encode_string (p, name);
   p = xdr_encode_create_state (p, mode);
@@ -481,13 +627,18 @@ netfs_attempt_rmdir (struct netcred *cred, struct node *np,
   
   /* Should we do the same sort of thing here as with attempt_unlink? */
 
-  p = nfs_initialize_rpc (NFSPROC_RMDIR, cred, 0, &rpcbuf, np, -1);
+  p = nfs_initialize_rpc (NFSPROC_RMDIR (protocol_version),
+			  cred, 0, &rpcbuf, np, -1);
   p = xdr_encode_fhandle (p, &np->nn->handle);
   p = xdr_encode_string (p, name);
   
   err = conduct_rpc (&rpcbuf, &p);
   if (!err)
-    err = nfs_error_trans (ntohl (*p++));
+    {
+      err = nfs_error_trans (ntohl (*p++));
+      if (protocol_version == 3)
+	p = process_wcc_stat (np, p, !err);
+    }
   
   free (rpcbuf);
   return err;
@@ -515,7 +666,8 @@ netfs_attempt_link (struct netcred *cred, struct node *dir,
     case POSSIBLE:
     case NOT_POSSIBLE:
       mutex_lock (&dir->lock);
-      p = nfs_initialize_rpc (NFSPROC_LINK, cred, 0, &rpcbuf, dir, -1);
+      p = nfs_initialize_rpc (NFSPROC_LINK (protocol_version),
+			      cred, 0, &rpcbuf, dir, -1);
       mutex_unlock (&dir->lock);
       
       mutex_lock (&np->lock);
@@ -537,7 +689,8 @@ netfs_attempt_link (struct netcred *cred, struct node *dir,
 
     case SYMLINK:
       mutex_lock (&dir->lock);
-      p = nfs_initialize_rpc (NFSPROC_SYMLINK, cred, 0, &rpcbuf, dir, -1);
+      p = nfs_initialize_rpc (NFSPROC_SYMLINK (protocol_version),
+			      cred, 0, &rpcbuf, dir, -1);
       p = xdr_encode_fhandle (p, &dir->nn->handle);
       mutex_unlock (&dir->lock);
       
@@ -552,78 +705,159 @@ netfs_attempt_link (struct netcred *cred, struct node *dir,
 	  return err;
 	}
       
-      p = xdr_encode_string (p, np->nn->transarg.name);
-      p = xdr_encode_sattr_stat (p, &np->nn_stat);
+      if (protocol_version == 2)
+	{
+	  p = xdr_encode_string (p, np->nn->transarg.name);
+	  p = xdr_encode_sattr_stat (p, &np->nn_stat);
+	}
+      else
+	{
+	  p = xdr_encode_sattr_stat (p, &np->nn_stat);
+	  p = xdr_encode_string (p, np->nn->transarg.name);
+	}	  
       mutex_unlock (&np->lock);
 
       mutex_lock (&dir->lock);
       err = conduct_rpc (&rpcbuf, &p);
-      if (!err)
-	err = nfs_error_trans (ntohl (*p++));
-      
       if (!err)
 	{
-	  /* NFSPROC_SYMLINK stupidly does not pass back an
-	     fhandle, so we have to fetch one now. */
-	  p = nfs_initialize_rpc (NFSPROC_LOOKUP, cred, 0, &rpcbuf, dir, -1);
-	  p = xdr_encode_fhandle (p, &dir->nn->handle);
-	  p = xdr_encode_string (p, name);
-	  
-	  err = conduct_rpc (&rpcbuf, &p);
-	  if (!err)
-	    err = nfs_error_trans (ntohl (*p++));
-
-	  mutex_unlock (&dir->lock);
-
-	  if (err)
-	    err = EGRATUITOUS;	/* damn */
-	  else
+	  err = nfs_error_trans (ntohl (*p++));
+      
+	  if (protocol_version == 2 && !err)
 	    {
-	      mutex_lock (&np->lock);
-	      recache_handle (np, p);
-	      p += NFS_FHSIZE / sizeof (int);
-	      register_fresh_stat (np, p);
-	      mutex_unlock (&np->lock);
+	      free (rpcbuf);
+
+	      /* NFSPROC_SYMLINK stupidly does not pass back an
+		 fhandle, so we have to fetch one now. */
+	      p = nfs_initialize_rpc (NFSPROC_LOOKUP (protocol_version),
+				      cred, 0, &rpcbuf, dir, -1);
+	      p = xdr_encode_fhandle (p, &dir->nn->handle);
+	      p = xdr_encode_string (p, name);
+	  
+	      mutex_unlock (&dir->lock);
+	  
+	      err = conduct_rpc (&rpcbuf, &p);
+	      if (!err)
+		err = nfs_error_trans (ntohl (*p++));
+	      if (!err)
+		{
+		  mutex_lock (&np->lock);
+		  p = recache_handle (p, np);
+		  p = process_returned_stat (np, p, 1);
+		  mutex_unlock (&np->lock);
+		}
+	      if (err)
+		err = EGRATUITOUS; /* damn */
 	    }
+	  else if (protocol_version == 3)
+	    {
+	      if (!err)
+		{
+		  mutex_unlock (&dir->lock);
+		  mutex_lock (&np->lock);
+		  p = recache_handle (p, np);
+		  p = process_returned_stat (np, p, 1);
+		  mutex_unlock (&np->lock);
+		  mutex_lock (&dir->lock);
+		}
+	      p = process_wcc_stat (dir, p, !err);
+	      mutex_unlock (&dir->lock);
+	    }
+	  else
+	    mutex_unlock (&dir->lock);
 	}
       else
 	mutex_unlock (&dir->lock);
+
+      free (rpcbuf);
       break;
       
     case CHRDEV:
     case BLKDEV:
     case FIFO:
     case SOCK:
-      mutex_lock (&dir->lock);
-      p = nfs_initialize_rpc (NFSPROC_CREATE, cred, 0, &rpcbuf, dir, -1);
-      p = xdr_encode_fhandle (p, &dir->nn->handle);
-      p = xdr_encode_string (p, name);
-      mutex_unlock (&dir->lock);
 
-      mutex_lock (&np->lock);
-      err = netfs_validate_stat (np, cred);
-      if (err)
+      if (protocol_version == 2)
 	{
+	  mutex_lock (&dir->lock);
+	  err = verify_nonexistent (cred, dir, name);
+	  if (err)
+	    return err;
+
+	  p = nfs_initialize_rpc (NFSPROC_CREATE (protocol_version),
+				  cred, 0, &rpcbuf, dir, -1);
+	  p = xdr_encode_fhandle (p, &dir->nn->handle);
+	  p = xdr_encode_string (p, name);
+	  mutex_unlock (&dir->lock);
+	  
+	  mutex_lock (&np->lock);
+	  err = netfs_validate_stat (np, cred);
+	  if (err)
+	    {
+	      mutex_unlock (&np->lock);
+	      free (rpcbuf);
+	      return err;
+	    }
+	  
+	  p = xdr_encode_sattr_stat (p, &np->nn_stat);
+	  mutex_unlock (&np->lock);
+
+	  mutex_lock (&dir->lock);
+	  err = conduct_rpc (&rpcbuf, &p);
+	  if (!err)
+	    err = nfs_error_trans (ntohl (*p++));
+	  mutex_unlock (&dir->lock);
+      
+	  mutex_lock (&np->lock);
+	  p = recache_handle (p, np);
+	  register_fresh_stat (np, p);
 	  mutex_unlock (&np->lock);
 	  free (rpcbuf);
-	  return err;
 	}
-      
-      p = xdr_encode_sattr_stat (p, &np->nn_stat);
-      mutex_unlock (&np->lock);
-
-      mutex_lock (&dir->lock);
-      err = conduct_rpc (&rpcbuf, &p);
-      if (!err)
-	err = nfs_error_trans (ntohl (*p++));
-      mutex_unlock (&dir->lock);
-      
-      mutex_lock (&np->lock);
-      recache_handle (np, p);
-      p += NFS_FHSIZE / sizeof (int);
-      register_fresh_stat (np, p);
-      mutex_unlock (&np->lock);
-
+      else
+	{
+	  mutex_lock (&dir->lock);
+	  p = nfs_initialize_rpc (NFS3PROC_MKNOD, cred, 0, &rpcbuf, dir, -1);
+	  p = xdr_encode_fhandle (p, &dir->nn->handle);
+	  p = xdr_encode_string (p, name);
+	  mutex_unlock (&dir->lock);
+	  
+	  mutex_lock (&np->lock);
+	  err = netfs_validate_stat (np, cred);
+	  if (err)
+	    {
+	      mutex_unlock (&np->lock);
+	      free (rpcbuf);
+	      return err;
+	    }
+	  *p++ = htonl (hurd_mode_to_nfs_type (np->nn_stat.st_mode));
+	  p = xdr_encode_sattr_stat (p, &np->nn_stat);
+	  if (np->nn->dtrans == BLKDEV || np->nn->dtrans == CHRDEV)
+	    {
+#define major(D) (((D)>>8) & 0xff)
+#define minor(D) ((D) & 0xff)
+	      *p++ = htonl (major (np->nn_stat.st_rdev));
+	      *p++ = htonl (minor (np->nn_stat.st_rdev));
+	    }
+	  mutex_unlock (&np->lock);
+	  
+	  err = conduct_rpc (&rpcbuf, &p);
+	  if (!err)
+	    {
+	      err = nfs_error_trans (ntohl (*p++));
+	      if (!err)
+		{
+		  mutex_lock (&np->lock);
+		  p = recache_handle (p, np);
+		  p = process_returned_stat (np, p, 1);
+		  mutex_unlock (&np->lock);
+		}
+	      mutex_lock (&dir->lock);
+	      p = process_wcc_stat (dir, p, !err);
+	      mutex_unlock (&dir->lock);
+	    }
+	  free (rpcbuf);
+	}
       break;
     }
 
@@ -704,22 +938,57 @@ netfs_attempt_create_file (struct netcred *cred, struct node *np,
   void *rpcbuf;
   error_t err;
 
-  p = nfs_initialize_rpc (NFSPROC_CREATE, cred, 0, &rpcbuf, np, -1);
+  /* RFC 1094 says that create is always exclusive.  But Sun doesn't
+     actually *implement* the spec.  No, of course not.  So we have to do
+     it for them.  */
+  if (protocol_version == 2)
+    {
+      err = verify_nonexistent (cred, np, name);
+      if (err)
+	return err;
+    }
+
+  p = nfs_initialize_rpc (NFSPROC_CREATE (protocol_version),
+			  cred, 0, &rpcbuf, np, -1);
   p = xdr_encode_fhandle (p, &np->nn->handle);
   p = xdr_encode_string (p, name);
-  p = xdr_encode_create_state (p, mode);
+  if (protocol_version == 3)
+    {
+      /* We happen to know this is where the XID is. */
+      int verf = *(int *)rpcbuf;
+      
+      *p++ = ntohl (EXCLUSIVE);
+      /* 8 byte verf */
+      *p++ = ntohl (verf);
+      p++;
+    }
+  else
+    p = xdr_encode_create_state (p, mode);
   
   err = conduct_rpc (&rpcbuf, &p);
-  if (!err)
-    err = nfs_error_trans (ntohl (*p++));
 
   mutex_unlock (&np->lock);
 
   if (!err)
     {
-      *newnp = lookup_fhandle (p);
-      p += NFS_FHSIZE / sizeof (int);
-      register_fresh_stat (*newnp, p);
+      err = nfs_error_trans (ntohl (*p++));
+      if (!err)
+	{
+	  p = lookup_fhandle (p, newnp);
+	  p = process_returned_stat (*newnp, p, 1);
+	}
+      if (err)
+	*newnp = 0;
+      if (protocol_version == 3)
+	{
+	  if (*newnp)
+	    mutex_unlock (&(*newnp)->lock);
+	  mutex_lock (&np->lock);
+	  p = process_wcc_stat (np, p, 1);
+	  mutex_unlock (&np->lock);
+	  if (*newnp)
+	    mutex_lock (&(*newnp)->lock);
+	}
     }
   else
     *newnp = 0;
@@ -789,13 +1058,18 @@ netfs_attempt_unlink (struct netcred *cred, struct node *dir,
   netfs_nput (np);
 
   mutex_lock (&dir->lock);
-  p = nfs_initialize_rpc (NFSPROC_REMOVE, cred, 0, &rpcbuf, dir, -1);
+  p = nfs_initialize_rpc (NFSPROC_REMOVE (protocol_version),
+			  cred, 0, &rpcbuf, dir, -1);
   p = xdr_encode_fhandle (p, &dir->nn->handle);
   p = xdr_encode_string (p, name);
   
   err = conduct_rpc (&rpcbuf, &p);
   if (!err)
-    err = nfs_error_trans (ntohl (*p++));
+    {
+      err = nfs_error_trans (ntohl (*p++));
+      if (protocol_version == 3)
+	p = process_wcc_stat (dir, p, !err);
+    }
   
   free (rpcbuf);
 
@@ -817,7 +1091,8 @@ netfs_attempt_rename (struct netcred *cred, struct node *fromdir,
     return EOPNOTSUPP;		/* XXX */
 
   mutex_lock (&fromdir->lock);
-  p = nfs_initialize_rpc (NFSPROC_RENAME, cred, 0, &rpcbuf, fromdir, -1);
+  p = nfs_initialize_rpc (NFSPROC_RENAME (protocol_version),
+			  cred, 0, &rpcbuf, fromdir, -1);
   p = xdr_encode_fhandle (p, &fromdir->nn->handle);
   p = xdr_encode_string (p, fromname);
   mutex_unlock (&fromdir->lock);
@@ -829,7 +1104,15 @@ netfs_attempt_rename (struct netcred *cred, struct node *fromdir,
   
   err = conduct_rpc (&rpcbuf, &p);
   if (!err)
-    err = nfs_error_trans (ntohl (*p++));
+    {
+      err = nfs_error_trans (ntohl (*p++));
+      if (protocol_version == 3)
+	{
+	  mutex_lock (&fromdir->lock);
+	  p = process_wcc_stat (fromdir, p, !err);
+	  p = process_wcc_stat (todir, p, !err);
+	}
+    }
   
   free (rpcbuf);
   return err;
@@ -851,15 +1134,19 @@ netfs_attempt_readlink (struct netcred *cred, struct node *np,
       return 0;
     }
   
-  p = nfs_initialize_rpc (NFSPROC_READLINK, cred, 0, &rpcbuf, np, -1);
+  p = nfs_initialize_rpc (NFSPROC_READLINK (protocol_version),
+			  cred, 0, &rpcbuf, np, -1);
   p = xdr_encode_fhandle (p, &np->nn->handle);
   
   err = conduct_rpc (&rpcbuf, &p);
   if (!err)
-    err = nfs_error_trans (ntohl (*p++));
-  
-  if (!err)
-    p = xdr_decode_string (p, buf);
+    {
+      err = nfs_error_trans (ntohl (*p++));
+      if (protocol_version == 3)
+	p = process_returned_stat (np, p, 0);
+      if (!err)
+	p = xdr_decode_string (p, buf);
+    }
 
   free (rpcbuf);
   return err;
@@ -915,11 +1202,13 @@ error_t
 netfs_check_open_permissions (struct netcred *cred, struct node *np,
 			      int flags, int newnode)
 {
+  int modes;
+  
   if ((flags & (O_READ|O_WRITE|O_EXEC)) == 0)
     return 0;
   
-  if ((flags & (O_READ|O_WRITE|O_EXEC))
-      == (flags & guess_mode_use (np, cred)))
+  netfs_report_access (cred, np, &modes);
+  if ((flags & (O_READ|O_WRITE|O_EXEC)) == (flags & modes))
     return 0;
   else
     return EACCES;
@@ -932,9 +1221,53 @@ netfs_report_access (struct netcred *cred,
 		     struct node *np,
 		     int *types)
 {
-  *types = guess_mode_use (np, cred);
-}
+  if (protocol_version == 2)
+    *types = guess_mode_use (np, cred);
+  else
+    {
+      int *p;
+      void *rpcbuf;
+      error_t err;
+      int ret;
+      int write_check, execute_check;
 
+      err = netfs_validate_stat (np, cred);
+      if (err)
+	goto fallback;
+      if (S_ISDIR (np->nn_stat.st_mode))
+	{
+	  write_check = ACCESS3_MODIFY | ACCESS3_DELETE | ACCESS3_EXTEND;
+	  execute_check = ACCESS3_LOOKUP;
+	}
+      else
+	{
+	  write_check = ACCESS3_MODIFY;
+	  execute_check = ACCESS3_EXECUTE;
+	}
+
+      p = nfs_initialize_rpc (NFS3PROC_ACCESS, cred, 0, &rpcbuf, np, -1);
+      p = xdr_encode_fhandle (p, &np->nn->handle);
+      *p++ = htonl (ACCESS3_READ | write_check | execute_check);
+      
+      err = conduct_rpc (&rpcbuf, &p);
+      if (!err)
+	{
+	  err = nfs_error_trans (ntohl (*p++));
+	  p = process_returned_stat (np, p, 0);
+	  if (!err)
+	    {
+	      ret = ntohl (*p++);
+	      *types = ((ret & ACCESS3_READ ? O_READ : 0)
+			| (ret & write_check ? O_WRITE : 0)
+			| (ret & execute_check ? O_EXEC : 0));
+	    }
+	  else
+	    /* fall back, sigh. */
+          fallback:
+	    *types = guess_mode_use (np, cred);
+	}
+    }
+}
 
 /* These definitions have unfortunate side effects, don't use them,
    clever though they are. */
@@ -1106,7 +1439,8 @@ fetch_directory (struct netcred *cred, struct node *dir,
   while (!eof)
     {
       /* Fetch new directory entries */
-      p = nfs_initialize_rpc (NFSPROC_READDIR, cred, 0, &rpcbuf, dir, -1);
+      p = nfs_initialize_rpc (NFSPROC_READDIR (protocol_version),
+			      cred, 0, &rpcbuf, dir, -1);
       p = xdr_encode_fhandle (p, &dir->nn->handle);
       *p++ = cookie;
       *p++ = ntohl (read_size);
