@@ -515,10 +515,9 @@ trivfs_S_file_chmod (struct trivfs_protid *cred,
 
 
 error_t
-trivfs_S_file_access (struct trivfs_protid *cred,
-		      mach_port_t reply,
-		      mach_msg_type_name_t reply_type,
-		      int *allowed)
+trivfs_S_file_check_access (struct trivfs_protid *cred,
+			    mach_port_t reply, mach_msg_type_name_t reply_type,
+			    int *allowed)
 {
   if (!cred)
     return EOPNOTSUPP;
@@ -742,6 +741,46 @@ trivfs_S_io_read (struct trivfs_protid *cred,
   return !*datalen && cancel ? EINTR : 0;
 }
 
+error_t
+trivfs_S_io_pathconf (struct trivfs_protid *cred,
+		      mach_port_t reply,
+		      mach_msg_type_name_t reply_type,
+		      int name,
+		      int *val)
+{
+  if (!cred)
+    return EOPNOTSUPP;
+  
+  switch (name)
+    {
+    case _PC_LINK_MAX:
+    case _PC_NAME_MAX:
+    case _PC_PATH_MAX:
+    case _PC_PIPE_BUF:
+    case _PC_NO_TRUNC:
+    default:
+      return io_pathconf (cred->realnode, name, val);
+      
+    case _PC_MAX_CANON:
+      *val = rawq->hiwat;
+      return 0;
+      
+    case _PC_MAX_INPUT:
+      *val = inputq->hiwat;
+      return 0;
+      
+    case _PC_CHOWN_RESTRICTED:
+      /* We implement this locally, remember... */
+      *val = 1;
+      return 0;
+      
+    case _PC_VDISABLE:
+      *val = _POSIX_VDISABLE;
+      return 0;
+    }
+}    
+
+
 error_t
 trivfs_S_io_readable (struct trivfs_protid *cred,
 		      mach_port_t reply,
@@ -756,7 +795,10 @@ trivfs_S_io_readable (struct trivfs_protid *cred,
 
   mutex_lock (&global_lock);
   if ((cred->po->openmodes & O_READ) == 0)
-    return EBADF;
+    {
+      mutex_unlock (&global_lock);
+      return EBADF;
+    }
   *amt = qsize (inputq);
   if (remote_input_mode && *amt)
     --*amt;
