@@ -31,6 +31,7 @@
 #include <netdb.h>
 #include <time.h>
 #include <assert.h>
+#include <version.h>
 
 #include <netinet/in.h>
 #include <arpa/inet.h>
@@ -50,7 +51,8 @@ exec_reauth (auth_t auth, int secure, int must_reauth,
 	     mach_port_t *ports, unsigned num_ports,
 	     mach_port_t *fds, unsigned num_fds);
 
-char *argp_program_version = "login 1.0 (GNU " HURD_RELEASE ")";
+char *argp_program_version = STANDARD_HURD_VERSION (login);
+     
 
 /* Defaults for various login parameters.  */
 char *default_args[] = {
@@ -108,6 +110,7 @@ static struct argp_option options[] =
 };
 static char *args_doc = "[USER [ARG...]]";
 static char *doc =
+"Exec a program with uids and/or the environment changed appropriately.\v"
 "To give args to the shell without specifying a user, use - for USER.\n"
 "Current login parameters include HOME, SHELL, USER, NAME, and ROOT.";
 
