@@ -30,9 +30,11 @@
 #include <idvec.h>
 #include <ps.h>
 #include <error.h>
+#include <version.h>
+
 #include "psout.h"
 
-char *argp_program_version = "ps 1.0 (GNU " HURD_RELEASE ")";
+char *argp_program_version = STANDARD_HURD_VERSION (ps);
 
 #define OA OPTION_ARG_OPTIONAL
 
@@ -75,7 +77,7 @@ static const struct argp_option options[] =
                                       " (which defaults to the sid of the"
                                       " current process)"},
   {"sid",        0,       0,      OPTION_ALIAS | OPTION_HIDDEN},
-  {"sort",       'S',	  "FIELD",0, "Sort the output with respect to FIELD,"
+  {"sort",       's',	  "FIELD",0, "Sort the output with respect to FIELD,"
                                      " backwards if FIELD is prefixed by `-'"},
   {"threads",    'T',     0,      0,  "Show the threads for each process"},
   {"tty",        't',     "TTY",  OA, "Only show processes with controlling"
@@ -90,10 +92,10 @@ static const struct argp_option options[] =
 };
 
 char *args_doc = "[PID...]";
-
-char *doc = "The USER, LID, PID, PGRP, and SID arguments may also be comma \
-separated lists.  The System V options -u and -g may be accessed with -O and \
---pgrp.";
+char *doc =
+"Show information about processes PID... (default all `interesting' processes)"
+"\vThe USER, LID, PID, PGRP, and SID arguments may also be comma separated"
+" lists.  The System V options -u and -g may be accessed with -O and -G."; 
 
 int 
 parse_enum(char *arg, char **choices, char *kind, int allow_mismatches)
